/*
All handling of mouse events is in this file.

Games choose what kind of mouse handling they want by setting their "mouseHandling" field
to one of:
  "drag+drop":
    Player may drag any card for which card.parentNode.mayTakeCard(card) returns true; when they
    drop the card each member of Game.dragDropTargets[] is examined to see if the card overlaps
    it, and if it can be legally dropped there.
    Right-click is used for "intelligent" moving of cards.
    Double-left-click is calls Game.sendToFoundations(..)

BaseCardGame.initialise() then sets the "mouseHandler" field of the game to the appropriate
member of MouseHandlers[].  A game could just set mouseHandler directly if it needs strange
mouse handling.
*/



function initMouseHandlers() {

}



var MouseHandlers = {};



MouseHandlers["pyramid"] = {
  reset: function() {},
  mouseUp: function(e) {},
  mouseDown: function(e) {},
  mouseMove: function(e) {},

  mouseClick: function(e) {
    if(e.button!==0) return;
    const act = this._mouseClick(e);
    // otherwise the final click of the game gets caught by the window.onclick
    // handler for the game-won message, instantly starting a new game :(
    e.stopPropagation();
    if(act && act instanceof Select) {
      act.perform();
    } else {
      gSelectionHighlighter.unhighlight();
      Game.selectedCard = null;
      if(act) doo(act);
    }
  },

  _mouseClick: function(e) {
    var t = e.target, x = e.pageX, y = e.pageY;

    if(t.isStock) return t.getActionForClick(null);
    var tp = t.parentNode;
    if(tp.isStock) return tp.getActionForClick(t);

    // a <flex/>, the highlighter, or something like that
    if(!t.isCard && !t.isAnyPile) {
      // if t is a spacer between two piles we change t to the pile on the row above
      t = t.previousSibling;
      var rpp = t ? t.rightParent : null;
      if(t && t.isAnyPile && rpp && (rpp.boxObject.y+rpp.boxObject.height > y)) {
        t = rpp;
      } else {
        // user is probably trying to dismiss the selection
        return null;
      }
    }

    // With pyramid layouts click often end up targetted at empty piles;
    // here we work out which card the user was probably trying to click
    if(t.isAnyPile) {
      if(!t.isPile) return null; // clicking on an empty foundation or waste pile should do nothing

      while(!t.hasChildNodes()) {
        var lp = t.leftParent, rp = t.rightParent;
        if(rp && x > rp.boxObject.x) t = rp;
        else if(lp && lp.boxObject.x+lp.boxObject.width > x) t = lp;
        // is it the pile directly above?
        else if(lp && lp.rightParent) t = lp.rightParent;
        // user is probably being stupid
        else return null;
        // are we too low down anyway?
        if(t.boxObject.y+t.boxObject.height < y) return null;
      }
      // we're interested in cards, not piles
      t = t.firstChild;
    }

    return t.parentNode.getActionForClick(t);
  },

  rightClick: function() {}
};
